@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003-2014 Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,10 +19,13 @@
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
+#include <string>
 #include <vector>
 
 #include "common.h"
+#include "user_filter.h"
 
+int verbosity = 0;
 
 namespace {
 
@@ -47,9 +50,11 @@ const F_entry F_table[] =
   { "numbers",        Filter::numbers },
   { "numbers_only",   Filter::numbers_only },
   { "same_height",    Filter::same_height },
+  { "text_block",     Filter::text_block },
   { "upper_num",      Filter::upper_num },
+  { "upper_num_mark", Filter::upper_num_mark },
   { "upper_num_only", Filter::upper_num_only },
-  { 0, (Filter::Type)0 }
+  { 0, Filter::user }				// not a valid name
   };
 
 struct T_entry
@@ -74,8 +79,6 @@ const T_entry T_table[] =
 } // end namespace
 
 
-int verbosity = 0;
-
 void Ocrad::internal_error( const char * const msg )
   {
   std::fprintf( stderr, "ocrad: internal error: %s\n", msg );
@@ -126,7 +129,7 @@ void Charset::show_error( const char * const program_name,
     std::fputs( "Valid charset names:", stderr );
     for( int i = 0; i < charsets; ++i )
       std::fprintf( stderr, "  %s", charset_name[i] );
-    std::fputs( "\n", stderr );
+    std::fputc( '\n', stderr );
     }
   }
 
@@ -155,12 +158,20 @@ void Transformation::show_error( const char * const program_name,
   }
 
 
+Control::~Control()
+  {
+  for( unsigned f = filters.size(); f > 0; --f )
+    if( filters[f-1].user_filterp )
+      delete filters[f-1].user_filterp;
+  }
+
+
 bool Control::add_filter( const char * const program_name,
                           const char * const name )
   {
   for( int i = 0; F_table[i].name != 0; ++i )
     if( std::strcmp( name, F_table[i].name ) == 0 )
-      { filters.push_back( F_table[i].type ); return true; }
+      { filters.push_back( Filter( F_table[i].type ) ); return true; }
   if( verbosity >= 0 )
     {
     if( name && std::strcmp( name, "help" ) != 0 )
@@ -168,12 +179,30 @@ bool Control::add_filter( const char * const program_name,
     std::fputs( "Valid filter names:", stderr );
     for( int i = 0; F_table[i].name != 0; ++i )
       std::fprintf( stderr, "  %s", F_table[i].name );
-    std::fputs( "\n", stderr );
+    std::fputc( '\n', stderr );
     }
   return false;
   }
 
 
+int Control::add_user_filter( const char * const program_name,
+                              const char * const file_name )
+  {
+  const User_filter * const user_filterp = new User_filter( file_name );
+  const int retval = user_filterp->retval();
+  if( retval == 0 )
+    filters.push_back( Filter( user_filterp ) );
+  else
+    {
+    if( verbosity >= 0 )
+      std::fprintf( stderr,"%s: user filter: %s\n",
+                    program_name, user_filterp->error().c_str() );
+    delete user_filterp;
+    }
+  return retval;
+  }
+
+
 bool Control::set_format( const char * const name )
   {
   if( std::strcmp( name, "byte" ) == 0 ) { utf8 = false; return true; }
