@@ -1,10 +1,9 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011,
-    2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -44,22 +43,6 @@ public:
   };
 
 
-class Filter
-  {
-public:
-  enum Type { none, letters, letters_only, numbers, numbers_only };
-private:
-  Type type_;
-
-public:
-  Filter() : type_( none ) {}
-  bool set( const char * const name );
-  Type type() const { return type_; }
-  void show_error( const char * const program_name,
-                   const char * const arg ) const;
-  };
-
-
 class Transformation
   {
 public:
@@ -77,10 +60,27 @@ public:
   };
 
 
+class User_filter;
+
+struct Filter
+  {
+  enum Type { letters, letters_only, numbers, numbers_only, same_height,
+              text_block, upper_num, upper_num_mark, upper_num_only, user };
+
+  const User_filter * user_filterp;
+  Type type;
+
+  explicit Filter( const User_filter * p )
+    : user_filterp( p ), type( user ) {}
+  explicit Filter( const Type t )
+    : user_filterp( 0 ), type( t ) {}
+  };
+
+
 struct Control
   {
   Charset charset;
-  Filter filter;
+  std::vector< Filter > filters;
   FILE * outfile, * exportfile;
   int debug_level;
   char filetype;
@@ -89,6 +89,10 @@ struct Control
   Control()
     : outfile( stdout ), exportfile( 0 ),
       debug_level( 0 ), filetype( '4' ), utf8( false ) {}
+  ~Control();
 
+  bool add_filter( const char * const program_name, const char * const name );
+  int add_user_filter( const char * const program_name,
+                       const char * const file_name );
   bool set_format( const char * const name );
   };
