@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003-2014 Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@
 */
 
 #include <cctype>
+#include <cerrno>
 #include <climits>
 #include <cstdio>
 #include <cstdlib>
@@ -29,9 +30,9 @@
 #include <string>
 #include <vector>
 #include <stdint.h>
-#if defined(__MSVCRT__) || defined(__OS2__) || defined(_MSC_VER)
-#include <fcntl.h>
 #include <unistd.h>
+#if defined(__MSVCRT__) || defined(__OS2__) || defined(__DJGPP__)
+#include <fcntl.h>
 #include <io.h>
 #endif
 
@@ -39,15 +40,15 @@
 #include "common.h"
 #include "rational.h"
 #include "rectangle.h"
+#include "user_filter.h"
 #include "page_image.h"
 #include "textpage.h"
 
 
 namespace {
 
-const char * const Program_name = "GNU Ocrad";
 const char * const program_name = "ocrad";
-const char * const program_year = "2014";
+const char * const program_year = "2019";
 const char * invocation_name = 0;
 
 struct Input_control
@@ -69,19 +70,14 @@ struct Input_control
 void show_error( const char * const msg, const int errcode = 0,
                  const bool help = false )
   {
-  if( verbosity >= 0 )
-    {
-    if( msg && msg[0] )
-      {
-      std::fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 )
-        std::fprintf( stderr, ": %s.", std::strerror( errcode ) );
-      std::fputs( "\n", stderr );
-      }
-    if( help )
-      std::fprintf( stderr, "Try '%s --help' for more information.\n",
-                    invocation_name );
-    }
+  if( verbosity < 0 ) return;
+  if( msg && msg[0] )
+    std::fprintf( stderr, "%s: %s%s%s\n", program_name, msg,
+                  ( errcode > 0 ) ? ": " : "",
+                  ( errcode > 0 ) ? std::strerror( errcode ) : "" );
+  if( help )
+    std::fprintf( stderr, "Try '%s --help' for more information.\n",
+                  invocation_name );
   }
 
 
@@ -125,32 +121,38 @@ void show_help()
                "UTF-8 formats. The pbm, pgm and ppm formats are collectively known as pnm.\n"
                "\nOcrad includes a layout analyser able to separate the columns or blocks\n"
                "of text normally found on printed pages.\n"
+               "\nFor best results the characters should be at least 20 pixels high. If\n"
+               "they are smaller, try the --scale option. Scanning the image at 300 dpi\n"
+               "usually produces a character size good enough for ocrad.\n"
+               "Merged, very bold or very light (broken) characters are normally not\n"
+               "recognized correctly. Try to avoid them.\n"
                "\nUsage: %s [options] [files]\n", invocation_name );
   std::printf( "\nOptions:\n"
-               "  -h, --help               display this help and exit\n"
-               "  -V, --version            output version information and exit\n"
-               "  -a, --append             append text to output file\n"
-               "  -c, --charset=<name>     try '--charset=help' for a list of names\n"
-               "  -e, --filter=<name>      try '--filter=help' for a list of names\n"
-               "  -f, --force              force overwrite of output file\n"
-               "  -F, --format=<fmt>       output format (byte, utf8)\n"
-               "  -i, --invert             invert image levels (white on black)\n"
-               "  -l, --layout             perform layout analysis\n"
-               "  -o, --output=<file>      place the output into <file>\n"
-               "  -q, --quiet              suppress all messages\n"
-               "  -s, --scale=[-]<n>       scale input image by [1/]<n>\n"
-               "  -t, --transform=<name>   try '--transform=help' for a list of names\n"
-               "  -T, --threshold=<n%%>     threshold for binarization (0-100%%)\n"
-               "  -u, --cut=<l,t,w,h>      cut input image by given rectangle\n"
-               "  -v, --verbose            be verbose\n"
-               "  -x, --export=<file>      export results in ORF format to <file>\n" );
+               "  -h, --help                display this help and exit\n"
+               "  -V, --version             output version information and exit\n"
+               "  -a, --append              append text to output file\n"
+               "  -c, --charset=<name>      try '--charset=help' for a list of names\n"
+               "  -e, --filter=<name>       try '--filter=help' for a list of names\n"
+               "  -E, --user-filter=<file>  user-defined filter, see manual for format\n"
+               "  -f, --force               force overwrite of output file\n"
+               "  -F, --format=<fmt>        output format (byte, utf8)\n"
+               "  -i, --invert              invert image levels (white on black)\n"
+               "  -l, --layout              perform layout analysis\n"
+               "  -o, --output=<file>       place the output into <file>\n"
+               "  -q, --quiet               suppress all messages\n"
+               "  -s, --scale=[-]<n>        scale input image by [1/]<n>\n"
+               "  -t, --transform=<name>    try '--transform=help' for a list of names\n"
+               "  -T, --threshold=<n%%>      threshold for binarization (0-100%%)\n"
+               "  -u, --cut=<l,t,w,h>       cut input image by given rectangle\n"
+               "  -v, --verbose             be verbose\n"
+               "  -x, --export=<file>       export results in ORF format to <file>\n" );
   if( verbosity >= 1 )
     {
-    std::printf( "  -1..6                    pnm output file type (debug)\n"
-                 "  -C, --copy               'copy' input to output (debug)\n"
-                 "  -D, --debug=<level>      (0-100) output intermediate data (debug)\n" );
+    std::printf( "  -1..6                     pnm output file type (debug)\n"
+                 "  -C, --copy                'copy' input to output (debug)\n"
+                 "  -D, --debug=<level>       (0-100) output intermediate data (debug)\n" );
     }
-  std::printf( "If no files are specified, ocrad reads the image from standard input.\n"
+  std::printf( "\nIf no files are specified, ocrad reads the image from standard input.\n"
                "If the -o option is not specified, ocrad sends text to standard output.\n"
                "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
                "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
@@ -184,8 +186,6 @@ int process_file( FILE * const infile, const char * const infile_name,
                   const Input_control & input_control,
                   const Control & control )
   {
-  if( verbosity >= 1 )
-    std::fprintf( stderr, "processing file '%s'\n", infile_name );
   try
     {
     Page_image page_image( infile, input_control.invert );
@@ -201,7 +201,7 @@ int process_file( FILE * const infile, const char * const infile_name,
       else
         {
         if( verbosity >= 1 )
-          std::fprintf( stderr, "file '%s' totally cut away\n", infile_name );
+          std::fprintf( stderr, "file '%s' totally cut away.\n", infile_name );
         return 1;
         }
       }
@@ -231,11 +231,42 @@ int process_file( FILE * const infile, const char * const infile_name,
       if( control.exportfile ) textpage.xprint( control );
       }
     }
-  catch( Page_image::Error e ) { show_error( e.msg ); return 2; }
-  if( verbosity >= 1 ) std::fputs( "\n", stderr );
+  catch( Page_image::Error & e ) { show_error( e.msg ); return 2; }
+  if( verbosity >= 1 ) std::fputc( '\n', stderr );
   return 0;
   }
 
+int process_full_file( FILE * const infile, const char * const infile_name,
+                       const Input_control & input_control,
+                       const Control & control )
+  {
+  if( verbosity >= 1 )
+    std::fprintf( stderr, "processing file '%s'\n", infile_name );
+  int retval = 0;
+  while( true )
+    {
+    const int tmp = process_file( infile, infile_name, input_control, control );
+    if( retval < tmp ) retval = tmp;
+    if( control.outfile ) std::fflush( control.outfile );
+    if( control.exportfile ) std::fflush( control.exportfile );
+    if( infile != stdin ) break;
+    if( tmp <= 1 )		// detect EOF
+      {
+      int ch;
+      do ch = std::fgetc( infile ); while( ch == 0 || std::isspace( ch ) );
+      std::ungetc( ch, infile );
+      }
+    if( tmp > 1 || std::feof( infile ) || std::ferror( infile ) ) break;
+    }
+  if( std::fclose( infile ) != 0 )
+    {
+    show_error( ( infile != stdin ) ? "Error closing input file" :
+                                      "Error closing stdin", errno );
+    if( retval < 1 ) retval = 1;
+    }
+  return retval;
+  }
+
 } // end namespace
 
 
@@ -255,37 +286,39 @@ int main( const int argc, const char * const argv[] )
 
   const Arg_parser::Option options[] =
     {
-    { '1', 0,           Arg_parser::no  },
-    { '2', 0,           Arg_parser::no  },
-    { '3', 0,           Arg_parser::no  },
-    { '4', 0,           Arg_parser::no  },
-    { '5', 0,           Arg_parser::no  },
-    { '6', 0,           Arg_parser::no  },
-    { 'a', "append",    Arg_parser::no  },
-    { 'c', "charset",   Arg_parser::yes },
-    { 'C', "copy",      Arg_parser::no  },
-    { 'D', "debug",     Arg_parser::yes },
-    { 'e', "filter",    Arg_parser::yes },
-    { 'f', "force",     Arg_parser::no  },
-    { 'F', "format",    Arg_parser::yes },
-    { 'h', "help",      Arg_parser::no  },
-    { 'i', "invert",    Arg_parser::no  },
-    { 'l', "layout",    Arg_parser::no  },
-    { 'o', "output",    Arg_parser::yes },
-    { 'q', "quiet",     Arg_parser::no  },
-    { 's', "scale",     Arg_parser::yes },
-    { 't', "transform", Arg_parser::yes },
-    { 'T', "threshold", Arg_parser::yes },
-    { 'u', "cut",       Arg_parser::yes },
-    { 'v', "verbose",   Arg_parser::no  },
-    { 'V', "version",   Arg_parser::no  },
-    { 'x', "export",    Arg_parser::yes },
-    {  0 , 0,           Arg_parser::no  } };
+    { '1', 0,             Arg_parser::no  },
+    { '2', 0,             Arg_parser::no  },
+    { '3', 0,             Arg_parser::no  },
+    { '4', 0,             Arg_parser::no  },
+    { '5', 0,             Arg_parser::no  },
+    { '6', 0,             Arg_parser::no  },
+    { 'a', "append",      Arg_parser::no  },
+    { 'c', "charset",     Arg_parser::yes },
+    { 'C', "copy",        Arg_parser::no  },
+    { 'D', "debug",       Arg_parser::yes },
+    { 'e', "filter",      Arg_parser::yes },
+    { 'E', "user-filter", Arg_parser::yes },
+    { 'f', "force",       Arg_parser::no  },
+    { 'F', "format",      Arg_parser::yes },
+    { 'h', "help",        Arg_parser::no  },
+    { 'i', "invert",      Arg_parser::no  },
+    { 'l', "layout",      Arg_parser::no  },
+    { 'o', "output",      Arg_parser::yes },
+    { 'q', "quiet",       Arg_parser::no  },
+    { 's', "scale",       Arg_parser::yes },
+    { 't', "transform",   Arg_parser::yes },
+    { 'T', "threshold",   Arg_parser::yes },
+    { 'u', "cut",         Arg_parser::yes },
+    { 'v', "verbose",     Arg_parser::no  },
+    { 'V', "version",     Arg_parser::no  },
+    { 'x', "export",      Arg_parser::yes },
+    {  0 , 0,             Arg_parser::no  } };
 
   const Arg_parser parser( argc, argv, options );
   if( parser.error().size() )				// bad option
     { show_error( parser.error().c_str(), 0, true ); return 1; }
 
+  int retval = 0;
   int argind = 0;
   for( ; argind < parser.arguments(); ++argind )
     {
@@ -308,6 +341,9 @@ int main( const int argc, const char * const argv[] )
       case 'D': control.debug_level = std::strtol( arg, 0, 0 ); break;
       case 'e': if( !control.add_filter( program_name, arg ) ) return 1;
                 break;
+      case 'E': retval = control.add_user_filter( program_name, arg );
+                if( retval != 0 ) return retval;
+                break;
       case 'f': force = true; break;
       case 'F': if( !control.set_format( arg ) )
                   { show_error( "bad output format.", 0, true ); return 1; }
@@ -331,9 +367,9 @@ int main( const int argc, const char * const argv[] )
       }
     } // end process options
 
-#if defined(__MSVCRT__) || defined(__OS2__) || defined(_MSC_VER)
-  setmode( fileno( stdin ), O_BINARY );
-  setmode( fileno( stdout ), O_BINARY );
+#if defined(__MSVCRT__) || defined(__OS2__) || defined(__DJGPP__)
+  setmode( STDIN_FILENO, O_BINARY );
+  setmode( STDOUT_FILENO, O_BINARY );
 #endif
 
   if( outfile_name && std::strcmp( outfile_name, "-" ) != 0 )
@@ -349,7 +385,7 @@ int main( const int argc, const char * const argv[] )
     if( !control.outfile )
       {
       if( verbosity >= 0 )
-        std::fprintf( stderr, "Cannot open %s\n", outfile_name );
+        std::fprintf( stderr, "Can't open '%s'\n", outfile_name );
       return 1;
       }
     }
@@ -364,49 +400,40 @@ int main( const int argc, const char * const argv[] )
       if( !control.exportfile )
         {
         if( verbosity >= 0 )
-          std::fprintf( stderr, "Cannot open %s\n", exportfile_name );
+          std::fprintf( stderr, "Can't open '%s'\n", exportfile_name );
         return 1;
         }
       }
     std::fprintf( control.exportfile,
-                  "# Ocr Results File. Created by %s version %s\n",
-                  Program_name, PROGVERSION );
+                  "# Ocr Results File. Created by GNU Ocrad version %s\n",
+                  PROGVERSION );
     }
 
   // process any remaining command line arguments (input files)
-  FILE * infile = (argind < parser.arguments()) ? 0 : stdin;
-  const char *infile_name = "-";
-  int retval = 0;
-  while( true )
+  bool stdin_used = false;
+  for( bool first = true; first || argind < parser.arguments(); first = false )
     {
-    while( infile != stdin )
+    const char * infile_name = ( argind < parser.arguments() ) ?
+                               parser.argument( argind++ ).c_str() : "-";
+    FILE * infile;
+    if( std::strcmp( infile_name, "-" ) == 0 )
       {
-      if( infile ) std::fclose( infile );
-      if( argind >= parser.arguments() ) { infile = 0; break; }
-      infile_name = parser.argument( argind++ ).c_str();
-      if( std::strcmp( infile_name, "-" ) == 0 ) infile = stdin;
-      else infile = std::fopen( infile_name, "rb" );
-      if( infile ) break;
-      if( verbosity >= 0 )
-        std::fprintf( stderr, "Cannot open %s\n", infile_name );
-      if( retval == 0 ) retval = 1;
+      if( stdin_used ) continue; else stdin_used = true;
+      infile = stdin;
       }
-    if( !infile ) break;
-
-    int tmp = process_file( infile, infile_name, input_control, control );
-    if( infile == stdin )
+    else
       {
-      if( tmp <= 1 )
+      infile = std::fopen( infile_name, "rb" );
+      if( !infile )
         {
-        int ch;
-        do ch = std::fgetc( infile ); while( std::isspace( ch ) );
-        std::ungetc( ch, infile );
+        if( verbosity >= 0 )
+          std::fprintf( stderr, "Can't open '%s'\n", infile_name );
+        if( retval < 1 ) retval = 1;
+        continue;
         }
-      if( tmp > 1 || std::feof( infile ) || std::ferror( infile ) ) infile = 0;
       }
-    if( tmp > retval ) retval = tmp;
-    if( control.outfile ) std::fflush( control.outfile );
-    if( control.exportfile ) std::fflush( control.exportfile );
+    const int tmp = process_full_file( infile, infile_name, input_control, control );
+    if( retval < tmp ) retval = tmp;
     }
   if( control.outfile ) std::fclose( control.outfile );
   if( control.exportfile ) std::fclose( control.exportfile );
