@@ -1,5 +1,5 @@
 /*  Ocradlib - Optical Character Recognition library
-    Copyright (C) 2009-2014 Antonio Diaz Diaz.
+    Copyright (C) 2009-2019 Antonio Diaz Diaz.
 
     This library is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -110,7 +110,7 @@ int OCRAD_set_image( OCRAD_Descriptor * const ocrdes,
     if( ocrdes->page_image ) delete ocrdes->page_image;
     ocrdes->page_image = page_image;
     }
-  catch( std::bad_alloc )
+  catch( std::bad_alloc & )
     { ocrdes->ocr_errno = OCRAD_mem_error; return -1; }
   return 0;
   }
@@ -136,7 +136,7 @@ int OCRAD_set_image_from_file( OCRAD_Descriptor * const ocrdes,
     if( ocrdes->page_image ) delete ocrdes->page_image;
     ocrdes->page_image = page_image;
     }
-  catch( std::bad_alloc )
+  catch( std::bad_alloc & )
     { ocrdes->ocr_errno = OCRAD_mem_error; retval = -1; }
   catch( ... )
     { ocrdes->ocr_errno = OCRAD_bad_argument; retval = -1; }
