@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003-2014 Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -449,10 +449,10 @@ bool Page_image::cut( const Rational ltwh[4] )
   const int t = absolute_pos( ltwh[1], top(), bottom() );
   if( t > re.top() ) { if( t < re.bottom() ) re.top( t ); else return false; }
 
-  const int r = l + absolute_pos( ltwh[2], left(), right() ) - 1;
+  const int r = l + absolute_pos( ltwh[2], 0, width() ) - 1;
   if( r < re.right() ) { if( r > re.left() ) re.right( r ); else return false; }
 
-  const int b = t + absolute_pos( ltwh[3], top(), bottom() ) - 1;
+  const int b = t + absolute_pos( ltwh[3], 0, height() ) - 1;
   if( b < re.bottom() ) { if( b > re.top() ) re.bottom( b ); else return false; }
 
   if( re.width() < 3 || re.height() < 3 ) return false;
