@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003-2014 Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,6 +32,8 @@ struct Csegment			// cartesian (one-dimensional) segment
     { return ( seg.valid() && left <= seg.left && seg.right <= right ); }
   bool includes( const int col ) const
     { return ( left <= col && col <= right ); }
+  bool overlaps( const Csegment & seg ) const
+    { return ( valid() && seg.valid() && left <= seg.right && right >= seg.left ); }
 
   int distance( const Csegment & seg ) const;
   int distance( const int col ) const;
