@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003-2014 Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@
 #include "rectangle.h"
 #include "track.h"
 #include "ucs.h"
+#include "user_filter.h"
 #include "bitmap.h"
 #include "blob.h"
 #include "character.h"
@@ -38,7 +39,7 @@
 
 namespace {
 
-// Return the character position >= first preceding a big gap or eol.
+// Returns the character position >= first preceding a big gap or eol.
 //
 int find_big_gap( const Textline & line, const int first,
                   const int space_width_limit )
@@ -127,6 +128,15 @@ Rectangle Textline::charbox( const Character & c ) const
   }
 
 
+bool Textline::is_key_character( const int i ) const
+  {
+  if( i < big_initials_ || i >= characters() )
+    Ocrad::internal_error( "is_key_character, index out of bounds." );
+  return ( cpv[i]->isalnum() && cpv[i]->guess( 0 ).code != 'J' &&
+           cpv[i]->height() < 2 * height() && 2 * cpv[i]->height() > height() );
+  }
+
+
 void Textline::delete_character( const int i )
   {
   if( i < 0 || i >= characters() )
@@ -284,7 +294,7 @@ int Textline::mean_vcenter() const
 void Textline::print( const Control & control ) const
   {
   for( int i = 0; i < characters(); ++i ) character( i ).print( control );
-  std::fputs( "\n", control.outfile );
+  std::fputc( '\n', control.outfile );
   }
 
 
@@ -297,8 +307,8 @@ void Textline::dprint( const Control & control, const bool graph,
     for( int i = 0; i < characters(); ++i )
       if( !character(i).maybe(' ') )
         hist.add_sample( character(i).height() );
-    std::fprintf( control.outfile, "mean height = %d, median height = %d, track segments = %d\n",
-                  mean_height(), hist.median(), segments() );
+    std::fprintf( control.outfile, "mean height = %d, median height = %d, track segments = %d, big initials = %d\n",
+                  mean_height(), hist.median(), segments(), big_initials_ );
     }
 
   for( int i = 0; i < characters(); ++i )
@@ -307,7 +317,7 @@ void Textline::dprint( const Control & control, const bool graph,
     if( i < big_initials_ ) c.dprint( control, c, graph, recursive );
     else c.dprint( control, charbox( c ), graph, recursive );
     }
-  std::fputs( "\n", control.outfile );
+  std::fputc( '\n', control.outfile );
   }
 
 
@@ -348,13 +358,6 @@ void Textline::recognize1( const Charset & charset ) const
 void Textline::apply_filter( const Filter::Type filter )
   {
   bool modified = false;
-  for( int i = characters() - 1; i >= 0; --i )
-    {
-    Character & c = character( i );
-    if( !c.guesses() ) continue;
-    c.apply_filter( filter );
-    if( !c.guesses() ) { delete_character( i ); modified = true; }
-    }
   if( filter == Filter::same_height )
     {
     Histogram hist;
@@ -364,12 +367,55 @@ void Textline::apply_filter( const Filter::Type filter )
     const int median_height = hist.median();
     for( int i = characters() - 1; i >= 0; --i )
       if( !character(i).maybe(' ') &&
-          !Ocrad::similar( character(i).height(), median_height, 13, 2 ) )
+          !Ocrad::similar( character(i).height(), median_height, 10, 2 ) )
         { delete_character( i ); modified = true; }
     }
-  if( modified )		// remove leadind/trailing/duplicate spaces
+  else
+    {
     for( int i = characters() - 1; i >= 0; --i )
-      if( character(i).maybe(' ') &&
-          ( i == 0 || i == characters() - 1 || character(i-1).maybe(' ') ) )
-        delete_character( i );
+      {
+      Character & c = character( i );
+      if( !c.guesses() ) continue;
+      c.apply_filter( filter );
+      if( !c.guesses() && filter != Filter::upper_num_mark )
+        { delete_character( i ); modified = true; }
+      }
+    if( filter == Filter::upper_num_mark )
+      join_broken_unrecognized_characters();
+    }
+  if( modified ) remove_leadind_trailing_duplicate_spaces();
+  }
+
+
+void Textline::apply_user_filter( const User_filter & user_filter )
+  {
+  bool modified = false;
+  for( int i = characters() - 1; i >= 0; --i )
+    {
+    Character & c = character( i );
+    if( !c.guesses() ) continue;
+    c.apply_user_filter( user_filter );
+    if( !c.guesses() && user_filter.discard() )
+      { delete_character( i ); modified = true; }
+    }
+  if( user_filter.mark() ) join_broken_unrecognized_characters();
+  if( modified ) remove_leadind_trailing_duplicate_spaces();
+  }
+
+
+void Textline::join_broken_unrecognized_characters()
+  {
+  for( int i = characters() - 1; i > 0; --i )
+    if( !character(i).guesses() &&
+        character(i).h_overlaps( character( i - 1 ) ) )
+      delete_character( i );
+  }
+
+
+void Textline::remove_leadind_trailing_duplicate_spaces()
+  {
+  for( int i = characters() - 1; i >= 0; --i )
+    if( character(i).maybe(' ') &&
+        ( i == 0 || i == characters() - 1 || character(i-1).maybe(' ') ) )
+      delete_character( i );
   }
