@@ -1,10 +1,9 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011
-    Antonio Diaz Diaz.
+    Copyright (C) 2003-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -21,7 +20,7 @@
 #include "ucs.h"
 
 
-int UCS::base_letter( const int code ) throw()
+int UCS::base_letter( const int code )
   {
   switch( code )
     {
@@ -48,12 +47,15 @@ int UCS::base_letter( const int code ) throw()
     case COCIRCU:
     case COTILDE:
     case CODIAER: return 'O';
-    case CSCEDI : return 'S';
+    case CSCEDI :
+    case CSCARON: return 'S';
     case CUGRAVE:
     case CUACUTE:
     case CUCIRCU:
     case CUDIAER: return 'U';
-    case CYACUTE: return 'Y';
+    case CYACUTE:
+    case CYDIAER: return 'Y';
+    case CZCARON: return 'Z';
     case SAGRAVE:
     case SAACUTE:
     case SACIRCU:
@@ -77,19 +79,21 @@ int UCS::base_letter( const int code ) throw()
     case SOCIRCU:
     case SOTILDE:
     case SODIAER: return 'o';
-    case SSCEDI : return 's';
+    case SSCEDI :
+    case SSCARON: return 's';
     case SUGRAVE:
     case SUACUTE:
     case SUCIRCU:
     case SUDIAER: return 'u';
     case SYACUTE:
     case SYDIAER: return 'y';
+    case SZCARON: return 'z';
     default:      return 0;
     }
   }
 
 
-int UCS::compose( const int letter, const int accent ) throw()
+int UCS::compose( const int letter, const int accent )
   {
   switch( letter )
     {
@@ -118,6 +122,8 @@ int UCS::compose( const int letter, const int accent ) throw()
               if( accent == '`' ) return CUGRAVE;
               if( accent == '^' ) return CUCIRCU;
               if( accent == ':' ) return CUDIAER; break;
+    case 'Y': if( accent == '\'') return CYACUTE;
+              if( accent == ':' ) return CYDIAER; break;
     case 'Z': return CZCARON;
     case 'a': if( accent == '\'') return SAACUTE;
               if( accent == '`' ) return SAGRAVE;
@@ -155,25 +161,32 @@ int UCS::compose( const int letter, const int accent ) throw()
   }
 
 
-bool UCS::isalnum( const int code ) throw()
+bool UCS::isalnum( const int code )
   {
   return ( UCS::isalpha( code ) || UCS::isdigit( code ) );
   }
 
 
-bool UCS::isalpha( const int code ) throw()
+bool UCS::isalpha( const int code )
   {
   return ( ( code < 128 && std::isalpha( code ) ) || base_letter( code ) );
   }
 
 
-bool UCS::isdigit( const int code ) throw()
+bool UCS::ishigh( const int code )
   {
-  return ( code <= '9' && code >= '0' );
+  if( isupper( code ) || isdigit( code ) ) return true;
+  switch( code )
+    {
+    case 'b': case 'd': case 'f': case 'g': case 'h': case 'i': case 'j':
+    case 'k': case 'l': case 'p': case 'q': case 't': case 'y': case '|':
+      return true;
+    default : return false;
+    }
   }
 
 
-bool UCS::islower( const int code ) throw()
+bool UCS::islower( const int code )
   {
   if( code < 128 && std::islower( code ) ) return true;
   const int base = base_letter( code );
@@ -181,7 +194,7 @@ bool UCS::islower( const int code ) throw()
   }
 
 
-bool UCS::islower_ambiguous( const int code ) throw()
+bool UCS::islower_ambiguous( const int code )
   {
   if( islower_small_ambiguous( code ) ) return true;
   switch( code )
@@ -197,7 +210,7 @@ bool UCS::islower_ambiguous( const int code ) throw()
   }
 
 
-bool UCS::islower_small( const int code ) throw()
+bool UCS::islower_small( const int code )
   {
   if( code >= 128 || !std::islower( code ) ) return false;
   switch( code )
@@ -210,7 +223,7 @@ bool UCS::islower_small( const int code ) throw()
   }
 
 
-bool UCS::islower_small_ambiguous( const int code ) throw()
+bool UCS::islower_small_ambiguous( const int code )
   {
   if( code >= 128 || !std::islower( code ) ) return false;
   switch( code )
@@ -222,13 +235,13 @@ bool UCS::islower_small_ambiguous( const int code ) throw()
   }
 
 
-bool UCS::isspace( const int code ) throw()
+bool UCS::isspace( const int code )
   {
-  return ( code < 128 && std::isspace( code ) );
+  return ( code < 128 && std::isspace( code ) ) || code == 0xA0;
   }
 
 
-bool UCS::isupper( const int code ) throw()
+bool UCS::isupper( const int code )
   {
   if( code < 128 && std::isupper( code ) ) return true;
   const int base = base_letter( code );
@@ -236,7 +249,18 @@ bool UCS::isupper( const int code ) throw()
   }
 
 
-bool UCS::isvowel( int code ) throw()
+bool UCS::isupper_normal_width( const int code )
+  {
+  if( code >= 128 || !std::isupper( code ) ) return false;
+  switch( code )
+    {
+    case 'I': case 'J': case 'L': case 'M': case 'Q': case 'W': return false;
+    default : return true;
+    }
+  }
+
+
+bool UCS::isvowel( int code )
   {
   if( code >= 128 ) code = base_letter( code );
   if( !code || !std::isalpha( code ) ) return false;
@@ -246,7 +270,7 @@ bool UCS::isvowel( int code ) throw()
   }
 
 
-unsigned char UCS::map_to_byte( const int code ) throw()
+unsigned char UCS::map_to_byte( const int code )
   {
   if( code < 0 ) return 0;
   if( code < 256 ) return code;
@@ -260,6 +284,7 @@ unsigned char UCS::map_to_byte( const int code ) throw()
     case SSCEDI : return 0xFE;
     case CSCARON: return 0xA6;
     case SSCARON: return 0xA8;
+    case CYDIAER: return 0xBE;
     case CZCARON: return 0xB4;
     case SZCARON: return 0xB8;
     case EURO   : return 0xA4;
@@ -268,7 +293,25 @@ unsigned char UCS::map_to_byte( const int code ) throw()
   }
 
 
-const char * UCS::ucs_to_utf8( const int code ) throw()
+int UCS::map_to_ucs( const unsigned char ch )
+  {
+  switch( ch )
+    {
+    case 0xA4: return EURO;
+    case 0xA6: return CSCARON;
+    case 0xA8: return SSCARON;
+    case 0xB4: return CZCARON;
+    case 0xB8: return SZCARON;
+    case 0xBC: return CLIGOE;
+    case 0xBD: return SLIGOE;
+    case 0xBE: return CYDIAER;
+    }
+  return ch;
+  }
+
+
+      // does not work for 'code' == 0
+const char * UCS::ucs_to_utf8( const int code )
   {
   static char s[7];
 
@@ -284,24 +327,25 @@ const char * UCS::ucs_to_utf8( const int code ) throw()
 
   s[i] = 0; --i;
   int d = 0;
-  for( ; i > 0; --i, d+=6 )
+  for( ; i > 0; --i, d += 6 )
     s[i] = 0x80 | ( ( code >> d ) & 0x3F );		// 10XX XXXX
   s[0] = mask | ( code >> d );
   return s;
   }
 
 
-int UCS::to_nearest_digit( const int code ) throw()
+int UCS::to_nearest_digit( const int code )
   {
   switch( code )
     {
+    case 'D':
     case 'O':
     case 'Q':
     case 'o':     return '0';
-    case '|':
     case 'I':
     case 'L':
     case 'l':
+    case '|':
     case SINODOT: return '1';
     case 'Z':
     case 'z':     return '2';
@@ -322,7 +366,7 @@ int UCS::to_nearest_digit( const int code ) throw()
   }
 
 
-int UCS::to_nearest_letter( const int code ) throw()
+int UCS::to_nearest_letter( const int code )
   {
   switch( code )
     {
@@ -340,7 +384,33 @@ int UCS::to_nearest_letter( const int code ) throw()
   }
 
 
-int UCS::toupper( const int code ) throw()
+int UCS::to_nearest_upper_num( const int code )
+  {
+  switch( code )
+    {
+    case '(':
+    case '[':     return 'C';
+    case 'l':
+    case '|':     return 'I';
+    case DEG:     return 'O';
+    case MICRO:   return 'U';
+    case POW1:
+    case SINODOT: return '1';
+    case POW2:    return '2';
+    case POW3:    return '3';
+    case 'q':     return '4';
+    case 'b':
+    case SOACUTE: return '6';
+    case '&':     return '8';
+    case 'g':
+    case MASCORD: return '9';
+    }
+  if( islower_ambiguous( code ) ) return toupper( code );
+  return code;
+  }
+
+
+int UCS::toupper( const int code )
   {
   if( code < 128 ) return std::toupper( code );
   switch( code )
@@ -368,11 +438,14 @@ int UCS::toupper( const int code ) throw()
     case SOTILDE: return COTILDE;
     case SODIAER: return CODIAER;
     case SSCEDI : return CSCEDI;
+    case SSCARON: return CSCARON;
     case SUGRAVE: return CUGRAVE;
     case SUACUTE: return CUACUTE;
     case SUCIRCU: return CUCIRCU;
     case SUDIAER: return CUDIAER;
     case SYACUTE: return CYACUTE;
+    case SYDIAER: return CYDIAER;
+    case SZCARON: return CZCARON;
     default:      return code;
     }
   }
